import threading
import time
from pyhtmlgui import Observable


class Counter(Observable):
    def __init__(self):
        super().__init__()
        self.value = 0
        self.active = True
        self.worker_thread = threading.Thread(target=self._worker_thread, daemon=True)
        self.worker_thread.start()

    def _worker_thread(self):
        while True:
            if self.active is True:
                self.set(self.value + 1)
            time.sleep(1)

    def start(self):
        self.active = True
        self.notify_observers()

    def stop(self):
        self.active = False
        self.notify_observers()

    def set(self, value):
        self.value = value
        self.notify_observers()

    def reset(self):
        self.set(0)



import sys
import os
import subprocess
import whichcraft
import webbrowser


class Browser:
    def __init__(self, browsername="default", executable=None):
        self.browsername = browsername
        self.executable = executable
        self.browserInstance = None
        if browsername == "default":
            self.browserInstance = BrowserDefault(self.executable)
        elif browsername == "chrome":
            self.browserInstance = BrowserChrome(self.executable)
        elif browsername == "electron":
            self.browserInstance = BrowserElectron(self.executable)
        else:
            raise Exception("Unknown mode '%s', use 'default' for system browser or 'chrome' or 'electron'")

    def open(self, browser_args, **kwargs) -> None:
        self.browserInstance.run(browser_args, **kwargs)


class BrowserDefault:
    def __init__(self, executable: str = None):
        self.executable = executable

    def run(self, browser_args, **kwargs) -> None:
        start_url = browser_args[0]
        if not start_url.startswith("http"):
            start_url = "http://" + start_url
        if self.executable is None:
            webbrowser.open(start_url)
        else:
            cmd = [self.executable, start_url]
            subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=sys.stderr, stdin=subprocess.PIPE)


class BrowserElectron:
    def __init__(self, executable: str) -> None:
        self.executable = executable
        if self.executable is None:
            self.executable = self._find_path()
        if self.executable is None:
            raise Exception("Electron executable could not be found")
        if not os.path.isfile(self.executable):
            raise Exception("Electron executable could not be found at '%s'" % self.executable)

    def run(self, browser_args, env=None, **kwargs) -> None:
        if env is None:
            env = os.environ.copy()
        cmd = [
                  self.executable,
                  '--no-sandbox',
                  '--disable-http-cache',
              ] + browser_args
        subprocess.Popen(cmd, stdout=sys.stdout, stderr=sys.stderr, stdin=subprocess.PIPE, env=env)

    @staticmethod
    def _find_path():
        if sys.platform in ['win32', 'win64']:
            # It doesn't work well passing the .bat file to Popen, so we get the actual .exe
            bat_path = whichcraft.which('electron')
            if bat_path is None:
                return None
            return os.path.join(bat_path, r'..\node_modules\electron\dist\electron.exe')
        elif sys.platform in ['darwin', 'linux']:
            return whichcraft.which('electron')
        else:
            return None


class BrowserChrome:
    def __init__(self, executable: str) -> None:
        self.executable = executable
        if self.executable is None:
            self.executable = self._find_path()
        if self.executable is None:
            raise Exception("Chrome executable could not be found")
        if not os.path.isfile(self.executable):
            raise Exception("Chrome executable could not be found at '%s'" % self.executable)

    def run(self, browser_args, env=None, **kwargs) -> None:
        start_url = browser_args[0]
        if not start_url.startswith("http"):
            start_url = "http://" + start_url
        cmd = [
            self.executable,
            start_url,
        ]
        subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=sys.stderr, stdin=subprocess.PIPE)

    def _find_path(self):
        if sys.platform in ['win32', 'win64']:
            return self._find_chrome_win()
        elif sys.platform == 'darwin':
            return self._find_chrome_mac()
        elif sys.platform.startswith('linux'):
            return self._find_chrome_linux()
        else:
            return None

    @staticmethod
    def _find_chrome_mac():
        default_dir = r'/Applications/Google Chrome.app/Contents/MacOS/Google Chrome'
        if os.path.exists(default_dir):
            return default_dir
        # use mdfind ci to locate Chrome in alternate locations and return the first one
        name = 'Google Chrome.app'
        alternate_dirs = [x for x in subprocess.check_output(["mdfind", name]).decode().split('\n') if x.endswith(name)]
        if len(alternate_dirs):
            return alternate_dirs[0] + '/Contents/MacOS/Google Chrome'
        return None

    @staticmethod
    def _find_chrome_linux():
        import whichcraft as wch
        chrome_names = ['chromium-browser',
                        'chromium',
                        'google-chrome',
                        'google-chrome-stable']

        for name in chrome_names:
            chrome = wch.which(name)
            if chrome is not None:
                return chrome
        return None

    @staticmethod
    def _find_chrome_win():
        import winreg as reg
        reg_path = r'SOFTWARE\Microsoft\Windows\CurrentVersion\App Paths\chrome.exe'

        for install_type in reg.HKEY_CURRENT_USER, reg.HKEY_LOCAL_MACHINE:
            try:
                reg_key = reg.OpenKey(install_type, reg_path, 0, reg.KEY_READ)
                chrome_path = reg.QueryValue(reg_key, None)
                reg_key.Close()
            except WindowsError:
                chrome_path = None
            else:
                break

        return chrome_path



import uuid
from pyhtmlgui import Observable, ObservableDict
from .counter import Counter


class CounterInDict(Observable):
    def __init__(self):
        super().__init__()
        self.counters = ObservableDict()

    def add_counter(self):
        name = ("%s" % uuid.uuid4()).split("-")[0]
        self.counters[name] = Counter()

    def remove_counter(self, name):
        del self.counters[name]



import typing
import weakref


class WeakFunctionReferences:
    def __init__(self):
        self.references = {}

    def add(self, function: typing.Callable):
        # noinspection PyUnresolvedReferences
        obj = function.__self__
        fname = function.__name__
        callback_id = self._create_callback_id(function)
        wr = weakref.ref(obj, self._create_delete_callback(callback_id))
        self.references[callback_id] = (wr, fname)
        return callback_id

    def remove(self, function: typing.Callable):
        callback_id = self._create_callback_id(function)
        del self.references[callback_id]

    def get(self, callback_id: int) -> typing.Callable:
        weak_obj, fname = self.references[callback_id]
        return getattr(weak_obj(), fname)

    def get_all(self):
        callback_ids = [c for c in self.references.keys()]
        for callback_id in callback_ids:
            if callback_id in self.references.keys():
                yield self.get(callback_id)

    def _create_delete_callback(self, callback_id: int) -> typing.Callable:
        # noinspection PyUnusedLocal
        def f(wr):
            del self.references[callback_id]
        return f

    @staticmethod
    def _create_callback_id(function: typing.Callable) -> int:
        # noinspection PyUnresolvedReferences
        obj = function.__self__
        callback_id = hash("%s%s" % (id(obj), function.__name__)) & 0xffffffffffff
        if callback_id < 0:
            callback_id = callback_id * -1
        return callback_id



from .observable import Observable


class ObservableDict(dict, Observable):

    def __init__(self, *args, **kwargs):
        Observable.__init__(self)
        dict.__init__(self, *args, **kwargs)

    def popitem(self):
        key_value = dict.popitem(self)
        self.notify_observers(action="popitem", key=key_value[0], item=key_value[1])
        return key_value

    def pop(self, key, default=None):
        value = dict.pop(self, key, default)
        self.notify_observers(action="pop", key=key, item=value)
        return value

    def __setitem__(self, key, value):
        dict.__setitem__(self, key, value)
        self.notify_observers(action="setitem", key=key, item=value)

    def __delitem__(self, key):
        value = dict.__getitem__(self, key)
        dict.__delitem__(self, key)
        self.notify_observers(action="delitem", key=key, item=value)

    def update(self, dict2, **kwargs):
        items = []
        for key, value in dict2.items():
            items.append([key, value])
        dict.update(self, dict2)
        self.notify_observers(action="update", items=items)

    def clear(self):
        dict.clear(self)
        self.notify_observers(action="clear")



from pyhtmlgui import PyHtmlView
from .countersInDictView import CountersInDictView
from .countersInListView import CountersInListView
from .counterView import CounterView
from .twoCountersView import TwoCountersView
from .animationView import AnimationView


class AppView(PyHtmlView):
    TEMPLATE_FILE = "appView.html"

    def __init__(self, subject, parent, **kwargs):
        super().__init__(subject, parent, **kwargs)
        self.counterView = CounterView(subject=subject.counter, parent=self)
        self.counterInListView = CountersInListView(subject=subject.countersInList, parent=self)
        self.counterInDictView = CountersInDictView(subject=subject.countersInDict, parent=self)
        self.twoCounters = TwoCountersView(subject=subject.twoCounters, parent=self)
        self.animation = AnimationView(subject=subject, parent=self)

    # these functions are automatically called by pyHtmlGui if they exist in the main View class
    def on_frontend_ready(self, nr_of_connections):
        print("on_frontend_ready" , nr_of_connections)

    def on_frontend_disconnected(self, nr_of_connections):
        print("on_frontend_disconnected", nr_of_connections)



from pyhtmlgui import PyHtmlView, ObservableDictView


class CountersInDictView(PyHtmlView):
    TEMPLATE_FILE = "countersInDictView.html"

    def __init__(self, subject, parent, **kwargs):
        super().__init__(subject, parent, **kwargs)
        self.dictView = ObservableDictView(subject.counters, self, CounterDictItemView, dom_element="tbody")

    def remove_counter(self, counter_id):
        self.subject.remove_counter(counter_id)


class CounterDictItemView(PyHtmlView):
    DOM_ELEMENT = "tr"
    TEMPLATE_STR = '''
    <td>{{pyview.element_key()}}</td>
    <td>{{pyview.subject.value}}</td>
    <td>
        <button {% if pyview.subject.active == true %}disabled{% endif %} onclick='pyview.subject.start();'>Start</button>
        <button {% if pyview.subject.active == false%}disabled{% endif %} onclick='pyview.subject.stop();'>Stop</button> 
        <button onclick='pyview.subject.reset();'>Reset</button> 
        <button onclick='pyview.parent.parent.remove_counter("{{pyview.element_key() }}");'>Delete</button> <br>
    </td>
    '''



from __future__ import annotations
import typing
import weakref
import traceback
import time
import random
import string
from markupsafe import Markup
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from pyhtmlgui.lib.observable import Observable
from pyhtmlgui.pyhtmlguiInstance import PyHtmlGuiInstance

CHARACTERS = list(string.ascii_lowercase + string.digits)


class PyHtmlView:
    TEMPLATE_STR  = None
    TEMPLATE_FILE = None
    DOM_ELEMENT = "div"
    DOM_ELEMENT_CLASS = ""
    DOM_ELEMENT_EXTRAS = ""

    def __init__(self,
                 subject,
                 parent:  typing.Union[PyHtmlView, PyHtmlGuiInstance],
                 **kwargs):

        self.uid = "pv%s" % ("".join(random.choices(CHARACTERS, k=16)))
        self.is_visible = False

        parent._add_child(self)
        self._subject_wref = weakref.ref(subject, self._on_subject_died)
        self._parent_wref = weakref.ref(parent, None)

        self._subject = None  # this is replace for a short time on render by the actual resolved object
        self._children = weakref.WeakSet()
        self._observables = []
        self.__last_rendered = None  # timestamp of last rendering, for debug only
        self.__was_rendered = False

        if type(parent) == PyHtmlGuiInstance:
            self._instance = parent
        else:
            self._instance = parent._instance

        # attach default observation event and detach because default componens is invisible until rendered
        if self._on_subject_updated is not None:
            try:
                self.add_observable(self.subject)
            except Exception:  # detach all will thow an exception if the event can not be attached
                print("object type '%s' can not be observed" % type(subject))

    @property
    def subject(self):
        if self._subject is not None:
            return self._subject
        return self._subject_wref()

    @property
    def parent(self):
        return self._parent_wref()

    def _on_subject_updated(self, source, **kwargs) -> None:
        self.update()

    # noinspection PyUnusedLocal
    def _on_subject_died(self, wr) -> None:
        self.delete()

    def render(self) -> typing.Union[str, None]:
        """
            Return object rendered to html string. This function should be called from inside the jinja templates.
            Direct usage is not needed, Returns html string rendered from template
        """
        html = self._inner_html()
        if html is None:
            return None
        self.__last_rendered = time.time()
        if self.DOM_ELEMENT is None:
            return Markup(html)
        else:
            cls = self.DOM_ELEMENT_CLASS
            if cls == "":   cls = self.__class__.__name__
            if cls is None: cls = ""
            if cls != "":   cls = 'class="%s"' % cls

            return Markup('<%(el)s %(cls)s id="%(uid)s" %(ex)s>%(html)s</%(el)s>' % {
                "el"  : self.DOM_ELEMENT,
                "cls" : cls,
                "uid" : self.uid,
                "ex"  : self.DOM_ELEMENT_EXTRAS,
                "html": html,
            })

    def update(self) -> None:
        """
        update rendered view in DOM, must be visible
        """
        if self.is_visible is True:
            html_content = self.render()
            if html_content is not None:  # object might have died, in that case don't render
                self.call_javascript("pyhtmlgui.replace_element", [self.uid, html_content], skip_results=True)
        else:
            raise Exception("Can't update invisible components")

    def insert_element(self, index: int, element: PyHtmlView) -> bool:
        """
        Insert a new element into DOM at index.
        This is used for example in ObservableDictView and ObservableListView to insert newly created elements.
        """
        html_content = element.render()
        if html_content is not None:  # object might have died, in that case don't render
            self.call_javascript("pyhtmlgui.insert_element", [self.uid, index, html_content], skip_results=True)
            return True
        return False

    def move_element(self, index: int, element: PyHtmlView) -> bool:
        """
        Move existing element to position index.
        This is used for example in ObservableDictView and ObservableListView to reorder elements without rerendering.
        """
        self.call_javascript("pyhtmlgui.move_element", [self.uid, index, element.uid], skip_results=True)
        return True

    def delete(self, remove_from_dom: bool = True) -> None:
        """
        Delete element, detach events, remove from parent, remove from frontend if element is visible

        :param remove_from_dom: If parent element calls delete, remove_from_dom is set to False,
                                because parent removes itself + all childen from dom automatically
        """
        if self.is_visible is True and remove_from_dom is True:
            self.call_javascript("pyhtmlgui.remove_element", [self.uid], skip_results=True)
        self.set_visible(False)
        try:
            self.parent._remove_child(self)
        except:
            pass
        self._children.clear()

    def set_visible(self, visible: bool) -> None:
        """
        Set component and childens visibility, components that are not visible get their events detached
        """

        if self.is_visible == visible:
            return

        if visible is False:
            self.is_visible = False
            for observable, target in self._observables:
                try:
                    observable().detach_observer(target())  # resolve weak references
                except Exception:
                    pass
            for child in self._children:
                try:
                    child.set_visible(False)
                except:
                    pass
        else:
            self.is_visible = True
            for observable, target in self._observables:
                try:
                    observable().attach_observer(target())  # resolve weak references
                except Exception:
                    pass

    def call_javascript(self, js_function_name, args=None, skip_results=False):
        """
        Call javascript function in frontend.
        :param js_function_name: Name of javascript function
        :param args: Arguments for js function
        :param skip_results: Don't receive results, give some slight performance inprovement because
                             we don't wait for results
        """
        return self._instance.call_javascript(js_function_name, args, skip_results)

    def eval_javascript(self, script, skip_results=False, **kwargs):
        """
        Run script in frontend.
        :param script: Javascript source code
        :param skip_results: Don't receive results, give some slight performance inprovement because
                             we don't wait for results
        :param kwargs: kwargs are passed to javascript as "arg" variable
        :return:
        """
        if self.is_visible is False:
            raise Exception("Can't javascript_call invisible components")
        return self._instance.call_javascript("pyhtmlgui.eval_script", [script, kwargs], skip_results=skip_results)

    def eval_javascript_electron(self, script, skip_results=False, **kwargs):
        if self.is_visible is False:
            raise Exception("Can't javascript_call invisible components")
        return self._instance.call_javascript("electron.eval_script", [script, kwargs], skip_results=skip_results)

    def _inner_html(self) -> typing.Union[str, None]:
        self._subject = self.subject  # receive hard reference to obj to it does not die on us while rendering
        if self._subject is None:  # Observed object died before render
            return None

        for child in self._children:
            try:
                child.__was_rendered = False
            except:
                pass

        if self.is_visible is False:
            self.set_visible(True)

        try:
            html = self._instance.get_template(self).render({"pyview": self})
        except Exception:
            tb = traceback.format_exc()
            msg = " Exception while rendering Template: %s\n" % self.__class__.__name__
            msg += " %s" % tb.replace("\n", "\n  ").strip()
            self.call_javascript("pyhtmlgui.debug_msg", [msg])
            html = msg
            print(msg)

        # set children that have not been rendered in last pass to invisible
        for child in self._children:
            try:
                if child.__was_rendered is False and child.is_visible is True:
                    child.set_visible(False)
            except:
                pass

        self.__was_rendered = True
        self._subject = None  # remove hard reference to subject, it may die now
        return html

    def _add_child(self, child: PyHtmlView) -> None:
        self._children.add(child)

    def _remove_child(self, child: PyHtmlView) -> None:
        try:
            self._children.remove(child)
        except: # ignore if weak ref faild to resolve
            pass

    def add_observable(self, subject: Observable, target_function: typing.Callable = None) -> None:
        try:
            if not callable(subject.attach_observer) or not callable(subject.detach_observer):
                raise Exception("object type '%s' can not be observed" % type(subject))
        except Exception:
            raise Exception("object type '%s' can not be observed" % type(subject))
        if target_function is None:
            target_function = self._on_subject_updated

        if type(subject) != weakref.ref:
            subject = weakref.ref(subject)
        if type(target_function) != weakref.WeakMethod:
            target_function = weakref.WeakMethod(target_function)
        self._observables.append([subject, target_function])

    def remove_observable(self, subject: Observable, target_function: typing.Callable = None) -> None:
        if target_function is None:
            target_function = self._on_subject_updated
        to_remove = []
        for e in self._observables:
            e_subject, e_target_function = e
            if subject == e_subject():
                if target_function is None or target_function == e_target_function():
                    to_remove.append(e)
        for e in to_remove:
            self._observables.remove(e)



from __future__ import annotations
import typing
import jinja2
import weakref
import traceback
import re
import os
import sys
import inspect
import json
import queue
import importlib
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from pyhtmlgui.pyhtmlgui import PyHtmlGui
    from pyhtmlgui.view.pyhtmlview import PyHtmlView
from .lib import WeakFunctionReferences


class PyHtmlGuiInstance:
    def __init__(self, parent: PyHtmlGui):
        self._parent = parent
        self._websocket_connections = []
        self._children = weakref.WeakSet()
        self._template_env = jinja2.Environment(loader=parent.template_loader, autoescape=jinja2.select_autoescape())
        self._template_cache = {}
        self._call_number = 0
        self._function_references = WeakFunctionReferences()
        self._template_env.globals['_create_py_function_reference'] = self._create_function_reference
        self._view = self._parent.view_class(parent.app_instance, self)

    @property
    def connections_count(self) -> int:
        return len(self._websocket_connections)

    def update(self) -> None:
        self.call_javascript("pyhtmlgui.update_element", ["pyHtmlGuiBody", self._view.render()], skip_results=True)

    def set_visible(self, visible: bool) -> None:
        """
        Set component and childens visibility, components that are not visible get their events detached
        """
        if visible is False:
            for child in self._children:
                if child.is_visible is True:
                    child.set_visible(False)
        if visible is True:
            self.update()

    def call_javascript(self, js_function_name: str, args: list = None, skip_results: bool = False):
        """
        Call javascript function in frontend.
        :param js_function_name: Name of javascript function
        :param args: Arguments for js function
        :param skip_results: Don't receive results, give some slight performance improvement if we don't wait for result
        """
        self._call_number += 1
        to_delete = self._call_number - 100
        # clean old function references, this is needed to results don't stay around if you don't
        for websocket_connection in self._websocket_connections:
            if to_delete in websocket_connection.javascript_call_result_objects:
                del websocket_connection.javascript_call_result_objects[to_delete]
            if to_delete in websocket_connection.javascript_call_result_queues:
                del websocket_connection.javascript_call_result_queues[to_delete]

        if args is None:
            args = []
        javascript_call_object = {'call': self._call_number, 'name': js_function_name, 'args': args}

        if skip_results is True:
            javascript_call_result = None
            javascript_call_object["skip_results"] = True
        else:
            javascript_call_result = JavascriptCallResult(self._call_number)

        data = json.dumps(javascript_call_object, default=lambda o: None)
        for websocket_connection in self._websocket_connections:
            if skip_results is False:
                websocket_connection.javascript_call_result_objects[self._call_number] = javascript_call_result
                javascript_call_result.add_call(websocket_connection)
            websocket_connection.send(data)
        return javascript_call_result

    def websocket_loop(self, websocket_connection) -> None:
        self._websocket_connections.append(websocket_connection)
        websocket_connection.loop(self._websocket_process_message)
        self._websocket_connections.remove(websocket_connection)
        if len(self._websocket_connections) == 0:
            self.set_visible(False)

    def _websocket_process_message(self, message, websocket_connection):
        if 'call' in message:
            function_name = "Function not found"
            args = "  "
            try:
                if message['name'] == "call_python_function_with_args":
                    functioncall_id = message['args'][0]
                    args = message['args'][1]
                    function = self._function_references.get(functioncall_id)
                    # noinspection PyUnresolvedReferences
                    function_name = "%s.%s" % (function.__self__.__class__.__name__, function.__name__)
                    return_val = function(*args)

                elif message['name'] == "call_python_function":
                    functioncall_id = message['args'][0]
                    function = self._function_references.get(functioncall_id)
                    # noinspection PyUnresolvedReferences
                    function_name = "%s.%s" % (function.__self__.__class__.__name__, function.__name__)
                    return_val = function()

                elif message['name'] == "python_bridge":
                    function_name = "%s.python_bridge" % self.__class__.__name__
                    return_val = None
                    if hasattr(self._view, "on_electron_message"):
                        self._view.on_electron_message(message)

                elif message['name'] == "frontend_ready":
                    function_name = "%s.frontend_ready" % self.__class__.__name__
                    self.update()
                    return_val = None
                    if hasattr(self._view, "on_frontend_ready"):
                        self._view.on_frontend_ready(len(self._websocket_connections))

                else:
                    return_val = None
                    print("unknown python function", message['name'])

            except Exception:
                tb = traceback.format_exc()
                msg = " Exception in: %s(%s)\n" % (function_name, ("%s" % args)[1:-1])
                msg += " %s" % tb.replace("\n", "\n  ").strip()
                self.call_javascript("pyhtmlgui.debug_msg", [msg],skip_results=True)
                print(msg)
                return_val = None

            if not ("skip_results" in message and message["skip_results"] is True):
                data = json.dumps({'return': message['call'], 'value': return_val}, default=lambda o: None)
                websocket_connection.send(data)

        elif 'return' in message:
            call_id = message['return']
            del message['return']  # remove internal id from result before passing to next level
            if call_id in websocket_connection.javascript_call_result_objects:
                js_call_result = websocket_connection.javascript_call_result_objects[call_id]
                js_call_result.result_received(websocket_connection, message)
        else:
            print('Invalid message received: ', message)

    def _create_function_reference(self, function: typing.Union[typing.Callable, jinja2.runtime.Undefined]) -> str:
        if type(function) == jinja2.runtime.Undefined:
            # noinspection PyProtectedMember
            raise Exception("Undefined python method in script: '%s'" % function._undefined_name)
        return self._function_references.add(function)

    def get_template(self, item: PyHtmlView, force_reload: bool = False):
        """
        Receive template associated to item
        :param item: The view Object
        :param force_reload: Force reloading of template
        :return:
        """
        if force_reload is False:
            try:
                return self._template_cache[item.__class__.__name__]
            except KeyError:
                pass

        if item.__class__.TEMPLATE_FILE is not None:  # load from file
            file_to_monitor = self._template_env.get_template(item.TEMPLATE_FILE).filename
            string_to_render = open(file_to_monitor, "r").read()
        else:  # load from class
            if self._parent.auto_reload is False:
                string_to_render = item.TEMPLATE_STR
                file_to_monitor = None
            else:
                module_name = item.__module__
                if module_name is None or module_name == str.__class__.__module__:
                    module_fullname = item.__class__.__name__  # Avoid reporting __builtin__
                else:
                    module_fullname = module_name + '.' + item.__class__.__name__

                try:
                    file_to_monitor = os.path.abspath(inspect.getfile(item.__class__))
                except Exception:  # in case its in main. this may be a bug in inspect
                    file_to_monitor = os.path.abspath(sys.argv[0])

                if module_name == "__main__":
                    name = os.path.splitext(os.path.basename(file_to_monitor))[0]
                    module = __import__(name)
                    importlib.reload(module)  # reload should work on non complex objects in __main__, but not for more
                    for comp in module_fullname.split(".")[1:]:
                        module = getattr(module, comp)
                else:
                    loader = importlib.machinery.SourceFileLoader(module_name, file_to_monitor)
                    # noinspection PyUnresolvedReferences
                    spec = importlib.util.spec_from_loader(loader.name, loader)
                    # noinspection PyUnresolvedReferences
                    module = importlib.util.module_from_spec(spec)
                    loader.exec_module(module)
                    module = getattr(module, module_fullname.split(".")[-1])
                string_to_render = module.TEMPLATE_STR

        if self._parent.auto_reload is True:
            self._parent.add_file_to_monitor(file_to_monitor, item.__class__.__name__)

        #  replace onclick="pyview.my_function(arg1,arg2)"
        #  with    onclick="pyhtmlgui.call({{_create_py_function_reference(pyview.my_function)}}, arg1, arg2)
        # this a a convinience function is user does not have to type the annoying stuff and functions look cleaner
        string_to_render = self._prepare_template(string_to_render)

        try:
            self._template_cache[item.__class__.__name__] = self._template_env.from_string(string_to_render)
        except Exception as e:
            msg = "Failed to load Template "
            if item.TEMPLATE_FILE is not None:
                msg += "from File '%s': " % item.TEMPLATE_FILE
            else:
                msg += "from Class '%s': " % item.__class__.__name__
            msg += " %s" % e
            raise Exception(msg)

        return self._template_cache[item.__class__.__name__]

    @staticmethod
    def _prepare_template(template: str) -> str:
        """
            Replace onclick="pyview.my_function(arg1,arg2)"
            with    onclick="pyhtmlgui.call({{_create_py_function_reference(pyview.my_function)}}, arg1, arg2)
        """
        parts = re.split('({{|}}|{%|%})', template)
        index = 0
        while index < len(parts):
            if parts[index] == "{{" or parts[index] == "{%" :
                parts[index] = "%s%s%s" % (parts[index] , parts[index+1], parts[index+2])
                del parts[index+ 1]
                del parts[index+ 1]
            index += 1

        new_parts = []
        for i, part in enumerate(parts):
            if part.startswith("{{") or part.startswith("{%") or part.find("pyview.") == -1:
                new_parts.append(part)
            else:
                # noinspection RegExpSingleCharAlternation
                subparts = re.split(r'(>| |\(|=|\"|\'|\n|\r|\t|;)(pyview.[a-zA-Z0-9_.]+\()', part)
                for x, subpart in enumerate(subparts):
                    if subpart.startswith("pyview."):
                        subpart = subpart.replace("(", ")}}, ", 1)
                        subparts[x] = "pyhtmlgui.call({{_create_py_function_reference(%s" % subpart
                for sp in subparts:
                    new_parts.append(sp)
        return "".join(new_parts).replace(r'\pyview.', 'pyview.')

    def clear_template_cache(self, classnames: str = None) -> None:
        if classnames is None:
            self._template_cache = {}
        else:
            for classname in classnames:
                try:
                    del self._template_cache[classname]
                except Exception:
                    pass

    def _add_child(self, child: PyHtmlView) -> None:
        self._children.add(child)

    def _remove_child(self, child: PyHtmlView) -> None:
        self._children.remove(child)


class JavascriptCallResult:
    def __init__(self, call_id):
        self.call_id = call_id
        self.websocket_connections = weakref.WeakSet()
        self._results_missing = 0
        self._callback = None

    def add_call(self, websocket_connection):
        websocket_connection.javascript_call_result_queues[self.call_id] = queue.Queue()
        self.websocket_connections.add(websocket_connection)
        self._results_missing += 1

    def result_received(self, websocket_connection, result):
        websocket_connection.javascript_call_result_queues[self.call_id].put(result)
        self._results_missing -= 1
        if self._results_missing == 0 and self._callback is not None:
            self._callback(self._collect_results())

    def _collect_results(self):
        results = []
        for websocket_connection in self.websocket_connections:
            results.append(websocket_connection.javascript_call_result_queues[self.call_id].get())
            del websocket_connection.javascript_call_result_queues[self.call_id]
            del websocket_connection.javascript_call_result_objects[self.call_id]
        errors = [result["error"] for result in results if "error" in result]
        if len(errors) > 0:
            msg = "%s of %s connected frontends returned an error\n" % (len(errors), len(results))
            msg += "\n".join(errors)
            raise Exception(msg)
        return [result["value"] for result in results]

    def __call__(self, callback: typing.Union[typing.Callable, None]  = None):
        if callback is None:
            return self._collect_results()
        else:
            if self._results_missing == 0:
                callback(self._collect_results())
            else:
                self._callback = callback



import threading
import time
from pyhtmlgui import PyHtmlView
import random


class AnimationView(PyHtmlView):
    TEMPLATE_FILE = "animationView.html"

    def __init__(self, subject, parent, **kwargs):
        super().__init__(subject, parent, **kwargs)
        self.balls = []
        self._t = None
        self.add_balls(20)
        self._framerate = 0

    def get_framerate(self):
        return round(self._framerate)

    def nr_of_balls(self):
        return len(self.balls)

    def set_visible(self, visible):
        # overwrite set_visible to start animation if element gets visible
        super().set_visible(visible)
        if visible is True and self._t is None:
            self._t = threading.Thread(target=self._loop, daemon=True)
            self._t.start()

    def add_balls(self, nr_of_balls = 1):
        for i in range(nr_of_balls):
            self.balls.append(BallView(self.subject, self))
        if self.is_visible:
            self.update()

    def remove_ball(self):
        if any(self.balls):
            del self.balls[0]
            if self.is_visible:
                self.update()

    def _loop(self):
        last_ts_fr = time.time()
        while self.is_visible:
            start = time.time()
            for ball in self.balls:
                ball.move()
            diff = time.time() - start
            sleep_time = (1/40) - diff # target ~40 fps
            if sleep_time > 0:
                time.sleep(sleep_time)

            now = time.time()
            diff = now - last_ts_fr
            last_ts_fr = now
            framerate = ((self._framerate*20) + (1 / diff))/21
            if round(framerate) != round(self._framerate): # prevent redrawing the animation by using js directly to display current frame rate
                js = 'document.getElementById("framerate").innerHTML = args.framerate+"";'
                self.eval_javascript(js, skip_results=True, framerate=round(framerate))
            self._framerate = framerate
        self._t = None


class BallView(PyHtmlView):
    DOM_ELEMENT = "li"
    TEMPLATE_STR = 'o'

    def __init__(self, subject, parent, **kwargs):
        super().__init__(subject, parent, **kwargs)
        self.position_x = random.randint(0, 100)
        self.position_y = random.randint(0, 100)
        self.direction_x = random.choice([-random.random()*2, random.random()*2])
        self.direction_y = random.choice([-random.random()*2, random.random()*2])
        self.color = "%s" % "".join(random.choices("456789abcdef", k=3))

    @property
    def DOM_ELEMENT_EXTRAS(self):
        return "style='color:#%s;display:block;position:absolute;left:%s%%;top:%s%%'" % (self.color, self.position_x, self.position_y)

    def move(self):
        self.position_x += self.direction_x
        if self.position_x >= 100 or self.position_x <= 0:
            self.direction_x *= -1

        self.position_y += self.direction_y
        if self.position_y >= 100 or self.position_y <= 0:
            self.direction_y *= -1

        if self.is_visible:
            # fast frontend update via direct js call
            js = 'item = document.getElementById(args.uid);'
            js += 'item.style.left = args.posx + "%";'
            js += 'item.style.top  = args.posy + "%";'
            self.eval_javascript(js, skip_results=True, uid=self.uid, posx=self.position_x, posy=self.position_y)
            # or even faster on the browser side, without eval by writing the js function in advance, see static/js/app.js
            #self.call_javascript("update_balls",[self.uid, self.position_x, self.position_y], skip_results=True)



import json
import os
import queue
import shutil
import sys
import threading
import time
import traceback
import uuid
import jinja2
import typing
from werkzeug.serving import make_server

from .lib import Browser
from .pyhtmlguiInstance import PyHtmlGuiInstance
from .view import PyHtmlView
import flask, flask_sock

class PyHtmlGui:
    def __init__(self,
                 app_instance     : object,
                 view_class       : typing.Type[PyHtmlView],
                 static_dir       : str             = "",
                 template_dir     : str             = "",
                 electron_app_dir : str             = None,
                 base_template    : str             = "pyHtmlGuiBase.html",
                 on_view_connected: typing.Callable = None,
                 on_view_disconnected: typing.Callable = None,
                 size             : tuple[int, int] = None,
                 position         : tuple[int, int] = None,
                 browser          : str             = "default",
                 executable       : str             = None,
                 listen_host      : str             = "127.0.0.1",
                 listen_port      : int             = 8000,
                 shared_secret    : str             = None,
                 auto_reload      : bool            = False,
                 single_instance  : bool            = True
                 ) -> None:
        """
        :param app_instance: Some object (eg. main program class instance), passed to view_class as obj on launch
        :param view_class: A class that Inherits from PyHtmlView
        :param static_dir: Static files, css, img go here
        :param template_dir: main.html and other html goes here
        :param electron_app_dir: in case we use electron, this is the electron.js file we launch,
                                 default file is in pyHtmlGui/assets/electron/main.py
        :param base_template: pyHtmlGuiBase in pyHtmlGui/assets/templates, or custom file in app templates dir
        :param size: window size
        :param on_view_connected: callback is called when a frontend connects via websocket,
                                        arguments passed: "nr of view instances", "nr of websocket connections"
        :param on_view_disconnected: callback is called when a frontend websocket is disconnected
                                        arguments passed: "nr of view instances", "nr of websocket connections"
        :param position: window position
        :param browser: default | chrome | electron
        :param executable: path to chrome/electron executable, if needed
        :param listen_host:
        :param listen_port:
        :param shared_secret: use "" to automatically generate a uid internally, None to disable token
        :param auto_reload: for development, monitor files and reload while app is running
        :param single_instance: create only one instance and share it between all connected websockets.
                                This is the default, so one instance of view_class is shared by all connected frontends
        """


        self.view_class = view_class
        self.app_instance = app_instance
        self.static_dir = os.path.abspath(static_dir)
        self.template_dir = os.path.abspath(template_dir)
        self.base_template = base_template
        self.size = size
        self.on_view_connected_callback = on_view_connected
        self.on_view_disconnected_callback = on_view_disconnected
        self.position = position
        self.browser = browser
        self.executable = executable
        self.electron_app_dir = electron_app_dir
        if shared_secret == "":
            self.shared_secret = "%s" % uuid.uuid4()
        elif shared_secret is None:
            self.shared_secret = None
        else:
            self.shared_secret = shared_secret
        self.listen_host = listen_host
        self.listen_port = listen_port
        self.auto_reload = auto_reload
        self.single_instance = single_instance

        if getattr(sys, 'frozen', False) is True:  # check if we are bundled by pyinstaller
            # noinspection PyUnresolvedReferences,PyProtectedMember
            lib_dir = os.path.join(sys._MEIPASS, "pyhtmlgui")
            self.auto_reload = False
        else:
            lib_dir = os.path.join(os.path.dirname(os.path.realpath(__file__)))

        self._template_dir = os.path.join(lib_dir, "assets", "templates")
        self._electron_dir = os.path.join(lib_dir, "assets", "electron")
        if self.electron_app_dir is None:  # use default internal app
            self.electron_app_dir = self._electron_dir
        else:
            if getattr(sys, 'frozen', False) is not True:
                # if we are not frozen, copy our internal lib to the electron target dir if it does not exist
                for f in ["pyhtmlgui.js", "main.js", "package.json"]:
                    lib_js_target = os.path.join(self.electron_app_dir, f)
                    lib_js_source = os.path.join(self._electron_dir, f)
                    if not os.path.exists(lib_js_target):
                        shutil.copyfile(lib_js_source, lib_js_target)

        if not os.path.exists(self.template_dir):
            raise Exception("Template dir '%s' not found" % self.template_dir)
        if not os.path.exists(self.static_dir):
            raise Exception("Static dir '%s' not found" % self.static_dir)

        self.template_loader = jinja2.FileSystemLoader(searchpath=[self._template_dir, self.template_dir])
        self._template_env = jinja2.Environment(loader=self.template_loader, autoescape=jinja2.select_autoescape())

        self._gui_instances = []

        self._token_cookie = "%s" % uuid.uuid4()
        self._token_csrf = "%s" % uuid.uuid4()

        self.flaskApp = flask.Flask("pyhtmlgui", static_folder=None, template_folder=None)
        self.flaskWebsocket =  flask_sock.Sock()

        self.flaskApp.route("/")(self._main_html)
        self.flaskApp.route("/static/<path:path>")(self._static)
        self.flaskWebsocket.route("/ws")(self._websocket)
        self.flaskWebsocket.init_app(self.flaskApp)

        self._server = WebsocketServerThread(self.flaskApp, self.listen_host, self.listen_port)
        self._browser = None

        self._file_monitoring = {}
        if self.auto_reload is True:
            t = threading.Thread(target=self._monitoring_thread, daemon=True)
            t.start()

    def start(self, show_frontend: bool = False, block: bool = True) -> None:
        if show_frontend is True:
            self.show()
        if block is True:
            self._server.run()
        else:
            self._server.start()

    def stop(self) -> None:
        # noinspection PyBroadException
        try:
            self._server.shutdown()
        except Exception:
            pass

    def show(self) -> None:
        env = None
        target_host = "127.0.0.1" if self.listen_host == "0.0.0.0" else self.listen_host

        if self.browser == "electron":
            args = [self.electron_app_dir, ]
            env = os.environ.copy()
            env.update({
                "PYHTMLGUI_HOST": target_host,
                "PYHTMLGUI_PORT": "%s" % self.listen_port,
                "PYHTMLGUI_SECRET": self.shared_secret,
            })
        else:
            args = ["%s:%s" % (target_host, self.listen_port), ]
            if self.shared_secret is not None:
                args[0] = "%s?token=%s" % (args[0], self.shared_secret)
        if self._browser is None:
            self._browser = Browser(self.browser, self.executable)
        self._browser.open(args, env=env)

    # /
    def _main_html(self):
        if self.shared_secret is not None and flask.request.args.get('token') != self.shared_secret:
            return flask.abort(403)
        template = self._template_env.get_template(self.base_template)
        response = flask.Response(template.render({
            "csrf_token": self._token_csrf,
            "start_size": json.dumps(self.size),
            "start_position": json.dumps(self.position),
        }))
        response.headers['Cache-Control'] = 'no-store'
        response.headers['Set-Cookie'] = 'token=%s' % self._token_cookie
        return response


    # /static/<path>
    def _static(self, path: str):
        if flask.request.cookies.get('token') != self._token_cookie:
            return flask.abort(403)
        response = flask.helpers.send_from_directory(
            self.static_dir, path, max_age=36000)

        if path.endswith(".js"):
            response.headers["Content-Type"] =  "application/javascript"
        elif path.endswith(".html"):
            response.headers["Content-Type"] =  "text/html"
        elif path.endswith(".css"):
            response.headers["Content-Type"] =  "text/css"
        elif path.endswith(".jpg") or path.endswith(".jpeg"):
            response.headers["Content-Type"] =  "image/jpeg"
        elif path.endswith(".png"):
            response.headers["Content-Type"] =  "image/png"
        elif path.endswith(".gif"):
            response.headers["Content-Type"] =  "image/gif"
        elif path.endswith(".webp"):
            response.headers["Content-Type"] =  "image/webp"
        elif path.endswith(".svg"):
            response.headers["Content-Type"] =  "image/svg"

        response.headers["Cache-Control"] = "public, max-age=36000"
        return response

    # /ws
    def _websocket(self, ws):
        if flask.request.cookies.get('token') != self._token_cookie or flask.request.args.get('token') != self._token_csrf:
            try:
                ws.close()
            except:
                pass
            return flask.abort(403)
        websocket_connection = WebsocketConnection(ws)
        # get/create Instance
        if self.single_instance is True:
            if len(self._gui_instances) == 0:
                self._gui_instances.append(PyHtmlGuiInstance(self))
            instance = self._gui_instances[0]
        else:
            instance = PyHtmlGuiInstance(self)
            self._gui_instances.append(instance)

        # onConnected Callback
        if self.on_view_connected_callback is not None:
            connections = sum([i.connections_count for i in self._gui_instances]) + 1
            # +1 because connection gets only added in loop later
            self.on_view_connected_callback(len(self._gui_instances), connections)

        #LOOP
        instance.websocket_loop(websocket_connection)  # loop while connected

        # release/remove Instance
        if instance.connections_count == 0:
            self._gui_instances.remove(instance)
        if hasattr(instance._view, "on_frontend_disconnected"):
            instance._view.on_frontend_disconnected(len(instance._websocket_connections))
        if self.on_view_disconnected_callback is not None:
            connections = sum([i.connections_count for i in self._gui_instances])
            self.on_view_disconnected_callback(len(self._gui_instances), connections)

    def add_file_to_monitor(self, file_to_monitor, class_name) -> None:
        if self.auto_reload is False:
            return
        if file_to_monitor not in self._file_monitoring:
            last_changed = os.path.getmtime(file_to_monitor)
            self._file_monitoring[file_to_monitor] = {
                "file_to_monitor": file_to_monitor,
                "last_changed": last_changed,
                "class_names": set(),
            }
        self._file_monitoring[file_to_monitor]["class_names"].add(class_name)

    def _monitoring_thread(self) -> None:
        while self.auto_reload is True:
            time.sleep(5)
            has_changed = False
            classed_to_reload = []
            for file_to_monitor in self._file_monitoring:
                data = self._file_monitoring[file_to_monitor]
                current_ts = os.path.getmtime(data["file_to_monitor"])
                if current_ts != data["last_changed"]:
                    self._file_monitoring[file_to_monitor]["last_changed"] = current_ts
                    classed_to_reload.extend(self._file_monitoring[file_to_monitor]["class_names"])
                    has_changed = True
            if has_changed is True:
                for instance in self._gui_instances:
                    instance.clear_template_cache(classed_to_reload)
                    try:
                        instance.update()
                    except Exception:
                        print("Failed to Update")
                        print(traceback.format_exc())


class WebsocketServerThread(threading.Thread):

    def __init__(self, app, host, port ):
        threading.Thread.__init__(self)
        self.server = make_server(host, port, app, threaded=True)
        self.ctx = app.app_context()
        self.ctx.push()

    def run(self):
        self.server.serve_forever()

    def shutdown(self):
        self.server.shutdown()


class WebsocketConnection:
    def __init__(self, ws):
        self.ws = ws
        self.javascript_call_result_queues = {}
        self.javascript_call_result_objects = {}
        self.active = True
        self.send_queue =queue.Queue()
        self._send_t = threading.Thread(target=self._send_loop, daemon=True)
        self._send_t.start()

    def loop(self, callback):
        while self.active is True:
            try:
                msg = self.ws.receive()
            except:
                msg = None
            if msg is None:
                break
            try:
                callback(json.loads(msg), self)
            except:
                continue
        try:
            self.ws.close()
        except:
            pass
        self.active = False
        if self.send_queue is not None:
            self.send_queue.put(None)

    def _send_loop(self):
        while self.active is True:
            message = self.send_queue.get()
            if message is None:
                break
            try:
                self.ws.send(message)
            except:
                pass
        try:
            self.ws.close()
        except:
            pass
        self.active = False
        self.send_queue = None

    def send(self, message):
        if self.send_queue is not None:
            self.send_queue.put(message)



from .weakFunctionReferences import WeakFunctionReferences
from .browser import Browser
from .observable import Observable
from .observableDict import ObservableDict
from .observableList import ObservableList



from pyhtmlgui import PyHtmlGui
import os
from src.app import App
from views.appView import AppView

SCRIPT_DIR = os.path.dirname(os.path.realpath(__file__))

if __name__ == "__main__":
    app = App()
    gui = PyHtmlGui(
        app_instance=app,
        view_class=AppView,
        template_dir=os.path.join(SCRIPT_DIR, "templates"),
        static_dir=os.path.join(SCRIPT_DIR, "static"),
        base_template="base.html",
        on_view_connected=app.on_view_connected,
        on_view_disconnected=app.on_view_disconnected,
        listen_port=8042,
        listen_host="127.0.0.1",
        auto_reload=True,
        single_instance=True,
        # Notice the animation and tab view in sync if you open the view in multiple browser windows
    )
    gui.start(show_frontend=True, block=True)



import random
from pyhtmlgui import Observable, ObservableList
from .counter import Counter


class CounterInList(Observable):
    def __init__(self):
        super().__init__()
        self.counters = ObservableList()

    def add_counter(self):
        self.counters.append(Counter())

    def remove_counter(self, index):
        del self.counters[index]

    def shuffle(self):
        random.shuffle(self.counters)



from pyhtmlgui import Observable
from .counter import Counter


class TwoCounters(Observable):
    def __init__(self):
        super().__init__()
        self.counter1 = Counter()
        self.counter2 = Counter()



from pyhtmlgui import PyHtmlView


class TwoCountersView(PyHtmlView):
    TEMPLATE_FILE = "twoCountersView.html"

    def __init__(self, subject, parent, **kwargs):
        super().__init__(subject, parent, **kwargs)
        self.counter1View = CounterPageView(subject.counter1, self, "Counter 1")
        self.counter2View = CounterPageView(subject.counter2, self, "Counter 2")
        self.current_counter = self.counter1View

    def show_page_1(self):
        if self.current_counter != self.counter1View:
            self.current_counter = self.counter1View
            self.update()

    def show_page_2(self):
        if self.current_counter != self.counter2View:
            self.current_counter = self.counter2View
            self.update()


class CounterPageView(PyHtmlView):
    TEMPLATE_STR = '''
        <h5>Page: {{ pyview.name }}</h5> <br>
        Value: {{ pyview.subject.value }} <br>
        Page update counter: {{ pyview.event_count }}  <br>  
    '''

    def __init__(self, subject, parent, name, **kwargs):
        super().__init__(subject, parent)
        self.name = name
        self.event_count = 0
        self.add_observable(subject, self.update_event_count)

    def update_event_count(self, *args, **kwargs) -> None:
        self.event_count += 1



from __future__ import annotations

import types
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from pyhtmlgui.lib.observableDict import ObservableDict
from .pyhtmlview import PyHtmlView
import typing
from threading import Lock


class ObservableDictView(PyHtmlView):
    TEMPLATE_STR = '''
        {% for item in pyview.get_items() %}
            {{ item.render()}}
        {% endfor %}
    '''

    def __init__(self,
                 subject        : ObservableDict,
                 parent         : PyHtmlView,
                 item_class     : type[PyHtmlView],
                 dom_element    : str             = PyHtmlView.DOM_ELEMENT,
                 sort_key       : typing.Callable = None,
                 sort_reverse   : bool            = False,
                 **kwargs):

        self._item_class = item_class
        self.DOM_ELEMENT = dom_element
        self._kwargs = kwargs
        self._wrapped_data = {}
        self._wrapped_data_lock = Lock()
        self.sort_key = sort_key
        self.sort_reverse = sort_reverse
        super().__init__(subject, parent)

    def set_visible(self, visible: bool) -> None:
        if self.is_visible == visible:  # not changed
            return
        self._wrapped_data_lock.acquire()
        super().set_visible(visible)
        self._wrapped_data = {}
        if self.is_visible is True:
            # is we were invisible, we might have missed add/delete events, so recreate our data wrapper
            for key, item in self.subject.items():
                self._wrapped_data[key] = self._create_item(item, key)
        self._wrapped_data_lock.release()

    def get_items(self) -> list:
        items = [item for key, item in self._wrapped_data.items()]
        if self.sort_key is None:
            return sorted(items, key=lambda x: x.element_key())
        else:
            return sorted(items, key=self.sort_key, reverse=self.sort_reverse)

    def _create_item(self, item, key):
        obj = self._item_class(item, self, **self._kwargs)
        obj.element_key = types.MethodType(lambda x: key, obj)
        return obj

    def _on_subject_updated(self, source, **kwargs) -> None:
        try:
            self._wrapped_data_lock.acquire()

            if kwargs["action"] == "setitem":
                if kwargs["key"] in self._wrapped_data:
                    self._wrapped_data[kwargs["key"]].delete()  # unrender
                obj = self._create_item(kwargs["item"], kwargs["key"])
                self._wrapped_data[kwargs["key"]] = obj
                index = list(self._wrapped_data.keys()).index(kwargs["key"])
                if self.insert_element(index, obj) is False:
                    del self._wrapped_data[kwargs["key"]]

            if kwargs["action"] == "update":
                for kv in kwargs["items"]:
                    key, item = kv
                    obj = self._create_item(item, kwargs["key"])
                    if key in self._wrapped_data:
                        self._wrapped_data[key].delete()  # unrender
                    self._wrapped_data[key] = obj
                    index = list(self._wrapped_data.keys()).index(kwargs["key"])
                    if self.insert_element(index, obj) is False:
                        del self._wrapped_data[kwargs["key"]]

            if kwargs["action"] in ["delitem", "pop", "popitem"]:
                self._wrapped_data[kwargs["key"]].delete()
                del self._wrapped_data[kwargs["key"]]

            if kwargs["action"] == "clear":
                for key, item in self._wrapped_data.items():
                    item.delete()
                self._wrapped_data.clear()
                self.update()

        finally:
            self._wrapped_data_lock.release()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()
print(setuptools.find_packages())

setuptools.setup(
    name="pyhtmlgui",
    version="2.15",
    author="Dirk Makerhafen",
    author_email="dirk@makerhafen.de",
    description="A Python library for building user interfaces",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/dirk-makerhafen/pyHtmlGui",
    #packages=setuptools.find_packages(),
    packages=setuptools.find_namespace_packages(),
    license='MIT',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    package_data={'pyhtmlgui': ['assets/electron/*', 'assets/templates/*']},
    include_package_data=True,
)



import os
import time
import json
from pyhtmlgui import PyHtmlGui, PyHtmlView, Observable

class App(Observable):
    pass

class AppView(PyHtmlView):
    TEMPLATE_STR = '''
        <button onclick="pyview.get_time().then(function(e){alert(e);})">Get system Time</button>   
        <button onclick="pyview.ping_electron()">Ping electron process</button>   
        <div id="electron_ping_result"></div>
        <button onclick="pyview.exit()">Exit App</button>   

    '''
    def ping_electron(self):
        self.call_javascript("electron.ping",[])(self._set_result)

    def _set_result(self, values):
        self.eval_javascript("document.getElementById('electron_ping_result').innerHTML = args.value", value = json.dumps(values))

    def get_time(self):
        return time.time()

    def exit(self):
        self.call_javascript("electron.exit",[], skip_results=True) # close electron app
        exit(0)

    def on_electron_message(self, message):
        print("electron message:", message)

if __name__ == "__main__":
    package_json = json.loads(open(os.path.join(os.path.dirname(__file__), "package.json"), "r").read()) # so electron and we use the same values
    gui = PyHtmlGui(
        browser          = "electron",
        app_instance     = App(),
        view_class       = AppView,
        listen_host      = package_json["PYHTMLGUI_HOST"],
        listen_port      = package_json["PYHTMLGUI_PORT"],
        shared_secret    = package_json["PYHTMLGUI_SECRET"],
    )
    gui.start(show_frontend=False, block=True)




import typing
from .weakFunctionReferences import WeakFunctionReferences


class Observable:
    def __init__(self):
        self._observers = WeakFunctionReferences()

    def attach_observer(self, target_function: typing.Callable) -> None:
        self._observers.add(target_function)

    def detach_observer(self, target_function: typing.Callable) -> None:
        self._observers.remove(target_function)

    def notify_observers(self, **kwargs) -> None:
        for target_function in self._observers.get_all():
            target_function(self, **kwargs)



from pyhtmlgui import Observable
from .counter import Counter
from .countersInDict import CounterInDict
from .countersInList import CounterInList
from .twoCounters import TwoCounters


class App(Observable):
    def __init__(self):
        super().__init__()
        self.counter = Counter()
        self.countersInDict = CounterInDict()
        self.countersInList = CounterInList()
        self.twoCounters = TwoCounters()

    def on_view_connected(self, nr_of_instances, nr_of_connections):
        print("View connected:", nr_of_instances, nr_of_connections)

    def on_view_disconnected(self, nr_of_instances, nr_of_connections):
        print("View disconnected:", nr_of_instances, nr_of_connections)
        if nr_of_instances == 0:
            print("No more frontends connected, exit now")
            exit(0)



import time, datetime, threading, random
from pyhtmlgui import PyHtmlGui, PyHtmlView, Observable


# App Logic
class CounterApp(Observable):
    def __init__(self):
        super().__init__()
        self.value = 0
        self.worker_thread = threading.Thread(target=self._worker_thread, daemon=True)
        self.worker_thread.start()

    def _worker_thread(self):
        while True:
            self.set_value(self.value + 1)
            time.sleep(1)

    def set_value(self, value):
        self.value = value
        self.notify_observers()


# View
class CounterAppView(PyHtmlView):
    TEMPLATE_STR = '''
        Current value: {{ pyview.subject.value }} <br>
        <button onclick='pyview.subject.set_value(0);'> Reset Counter </button> <br><br>
        <button onclick="pyview.get_time().then(function(e){alert(e);})"> Get System Time </button>
        <button onclick="pyview.call_js_from_python()"> Click and watch python console </button>
        <script>
            // script tags are executed every time this view element was rendered or updated
            document.getElementById("{{pyview.uid}}").style.backgroundColor = '#'+Math.floor(Math.random()*16777215).toString(16);
        </script>
    '''

    def get_time(self):
        return "It is now: %s" % datetime.datetime.now()

    def call_js_from_python(self):
        resultsHandler = self.eval_javascript(
            script='return document.getElementById(args.uid).style.backgroundColor;',
            uid=self.uid)
        resultsHandler(callback=lambda results: print(results))
        # results = resultsHandler() #synchronous, would break eventloop here if used in a function that in itself is called from javascript
        # note multiple results if multiple frontends are connected AND PyHtmlGui.single_instance is True (the default)


# Main
if __name__ == "__main__":
    gui = PyHtmlGui(
        app_instance = CounterApp(),
        view_class   = CounterAppView,
        auto_reload  = True, # edit templates while frontend is active!
    )
    gui.start(show_frontend=True, block=True)
